@@ -1,10 +1,11 @@
 import os
 import sqlite3
 import stat
-from typing import Optional
+import time
+from typing import Any, Callable, Dict, Optional, Union
 
 from jupyter_core.paths import jupyter_data_dir
-from traitlets import TraitError, Unicode, validate
+from traitlets import Int, TraitError, Unicode, validate
 from traitlets.config.configurable import LoggingConfigurable
 
 
@@ -36,14 +37,166 @@ def log(log_before, log_after):
     return decorator
 
 
-class FileIdManager(LoggingConfigurable):
+class BaseFileIdManager(LoggingConfigurable):
     """
-    Manager that supports tracking files across their lifetime by associating
-    each with a unique file ID, which is maintained across filesystem operations.
+    Base class for File ID manager implementations. All File ID
+    managers should inherit from this class.
+    """
+
+    root_dir = Unicode(
+        help=("The root directory being served by Jupyter server. Must be an absolute path."),
+        config=False,
+    )
+
+    db_path = Unicode(
+        default_value=default_db_path,
+        help=(
+            "The path of the DB file used by File ID manager implementations. "
+            "Defaults to `jupyter_data_dir()/file_id_manager.db`."
+        ),
+        config=True,
+    )
+
+    @validate("root_dir", "db_path")
+    def _validate_abspath_traits(self, proposal):
+        if proposal["value"] is None:
+            raise TraitError(f"FileIdManager : {proposal['trait'].name} must not be None")
+        if not os.path.isabs(proposal["value"]):
+            raise TraitError(f"FileIdManager : {proposal['trait'].name} must be an absolute path")
+        return proposal["value"]
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    def index(self, path: str) -> Union[int, str, None]:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def get_id(self, path: str) -> Union[int, str, None]:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def get_path(self, id: Union[int, str]) -> Union[int, str, None]:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def move(self, old_path: str, new_path: str) -> Union[int, str, None]:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def copy(self, from_path: str, to_path: str) -> Union[int, str, None]:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def delete(self, path: str) -> None:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def save(self, path: str) -> None:
+        raise NotImplementedError("must be implemented by subclass")
+
+    def get_handlers_by_action(self) -> Dict[str, Optional[Callable[[Dict[str, Any]], Any]]]:
+        """Returns a dictionary whose keys are contents manager event actions
+        and whose values are callables invoked upon receipt of an event of the
+        same action. The callable accepts the body of the event as its only
+        argument. To ignore an event action, set the value to `None`."""
+        raise NotImplementedError("must be implemented by subclass")
+
+
+class ArbitraryFileIdManager(BaseFileIdManager):
+    """
+    File ID manager that works on arbitrary filesystems. Each file is assigned a
+    unique ID. The path is only updated upon calling `move()`, `copy()`, or
+    `delete()`, e.g. upon receipt of contents manager events emitted by Jupyter
+    Server 2.
+    """
+
+    def __init__(self, *args, **kwargs):
+        # pass args and kwargs to parent Configurable
+        super().__init__(*args, **kwargs)
+        # initialize instance attrs
+        self._update_cursor = False
+        # initialize connection with db
+        self.log.info(f"ArbitraryFileIdManager : Configured root dir: {self.root_dir}")
+        self.log.info(f"ArbitraryFileIdManager : Configured database path: {self.db_path}")
+        self.con = sqlite3.connect(self.db_path)
+        self.log.info("ArbitraryFileIdManager : Successfully connected to database file.")
+        self.log.info("ArbitraryFileIdManager : Creating File ID tables and indices.")
+        # do not allow reads to block writes. required when using multiple processes
+        self.con.execute("PRAGMA journal_mode = WAL")
+        self.con.execute(
+            "CREATE TABLE IF NOT EXISTS Files("
+            "id INTEGER PRIMARY KEY AUTOINCREMENT, "
+            "path TEXT NOT NULL UNIQUE"
+            ")"
+        )
+        self.con.execute("CREATE INDEX IF NOT EXISTS ix_Files_path ON Files (path)")
+        self.con.commit()
+
+    def index(self, path: str) -> int:
+        row = self.con.execute("SELECT id FROM Files WHERE path = ?", (path,)).fetchone()
+        existing_id = row and row[0]
+
+        if existing_id:
+            return existing_id
+
+        # create new record
+        cursor = self.con.execute("INSERT INTO Files (path) VALUES (?)", (path,))
+        self.con.commit()
+        return cursor.lastrowid  # type:ignore
+
+    def get_id(self, path: str) -> Optional[int]:
+        row = self.con.execute("SELECT id FROM Files WHERE path = ?", (path,)).fetchone()
+        return row and row[0]
+
+    def get_path(self, id: Union[int, str]) -> Optional[int]:
+        row = self.con.execute("SELECT path FROM Files WHERE id = ?", (id,)).fetchone()
+        return row and row[0]
+
+    def move(self, old_path: str, new_path: str) -> None:
+        row = self.con.execute("SELECT id FROM Files WHERE path = ?", (old_path,)).fetchone()
+        id = row and row[0]
+
+        if id:
+            self.con.execute("UPDATE Files SET path = ? WHERE path = ?", (new_path, old_path))
+        else:
+            cursor = self.con.execute("INSERT INTO Files (path) VALUES (?)", (new_path,))
+            id = cursor.lastrowid
+
+        self.con.commit()
+        return id
+
+    def copy(self, from_path: str, to_path: str) -> Optional[int]:
+        cursor = self.con.execute("INSERT INTO Files (path) VALUES (?)", (to_path,))
+        self.con.commit()
+        return cursor.lastrowid
+
+    def delete(self, path: str) -> None:
+        self.con.execute("DELETE FROM Files WHERE path = ?", (path,))
+        self.con.commit()
+
+    def save(self, path: str) -> None:
+        return
+
+    def get_handlers_by_action(self) -> Dict[str, Optional[Callable[[Dict[str, Any]], Any]]]:
+        return {
+            "get": None,
+            "save": None,
+            "rename": lambda data: self.move(data["source_path"], data["path"]),
+            "copy": lambda data: self.copy(data["source_path"], data["path"]),
+            "delete": lambda data: self.delete(data["path"]),
+        }
+
+    def __del__(self):
+        """Cleans up `ArbitraryFileIdManager` by committing any pending
+        transactions and closing the connection."""
+        if hasattr(self, "con"):
+            self.con.commit()
+            self.con.close()
+
+
+class LocalFileIdManager(BaseFileIdManager):
+    """
+    File ID manager that supports tracking files in local filesystems by
+    associating each with a unique file ID, which is maintained across
+    filesystem operations.
 
     Notes
     -----
-
     All private helper methods prefixed with an underscore (except `__init__()`)
     do NOT commit their SQL statements in a transaction via `self.con.commit()`.
     This responsibility is delegated to the public method calling them to
@@ -52,15 +205,13 @@ class FileIdManager(LoggingConfigurable):
     performed during a method's procedure body.
     """
 
-    root_dir = Unicode(
-        help=("The root being served by Jupyter server. Must be an absolute path."), config=True
-    )
-
-    db_path = Unicode(
-        default_value=default_db_path,
+    autosync_interval = Int(
+        default_value=5,
         help=(
-            "The path of the DB file used by `FileIdManager`. "
-            "Defaults to `jupyter_data_dir()/file_id_manager.db`."
+            "How often to automatically invoke `sync_all()` when calling `get_path()`, in seconds. "
+            "Set to 0 to force `sync_all()` to always be called when calling `get_path()`. "
+            "Set to a negative value to disable autosync for `get_path()`."
+            "Defaults to 5."
         ),
         config=True,
     )
@@ -70,12 +221,13 @@ class FileIdManager(LoggingConfigurable):
         super().__init__(*args, **kwargs)
         # initialize instance attrs
         self._update_cursor = False
+        self._last_sync = 0.0
         # initialize connection with db
-        self.log.info(f"FileIdManager : Configured root dir: {self.root_dir}")
-        self.log.info(f"FileIdManager : Configured database path: {self.db_path}")
+        self.log.info(f"LocalFileIdManager : Configured root dir: {self.root_dir}")
+        self.log.info(f"LocalFileIdManager : Configured database path: {self.db_path}")
         self.con = sqlite3.connect(self.db_path)
-        self.log.info("FileIdManager : Successfully connected to database file.")
-        self.log.info("FileIdManager : Creating File ID tables and indices.")
+        self.log.info("LocalFileIdManager : Successfully connected to database file.")
+        self.log.info("LocalFileIdManager : Creating File ID tables and indices.")
         # do not allow reads to block writes. required when using multiple processes
         self.con.execute("PRAGMA journal_mode = WAL")
         self.con.execute(
@@ -96,14 +248,6 @@ class FileIdManager(LoggingConfigurable):
         self.con.execute("CREATE INDEX IF NOT EXISTS ix_Files_is_dir ON Files (is_dir)")
         self.con.commit()
 
-    @validate("root_dir", "db_path")
-    def _validate_abspath_traits(self, proposal):
-        if proposal["value"] is None:
-            raise TraitError(f"FileIdManager : {proposal['trait'].name} must not be None")
-        if not os.path.isabs(proposal["value"]):
-            raise TraitError(f"FileIdManager : {proposal['trait'].name} must be an absolute path")
-        return self._normalize_path(proposal["value"])
-
     def _index_all(self):
         """Recursively indexes all directories under the server root."""
         self._index_dir_recursively(self.root_dir, self._stat(self.root_dir))
@@ -140,6 +284,7 @@ class FileIdManager(LoggingConfigurable):
         _sync_file(). Hence the cursor needs to be redefined if
         self._update_cursor is set to True by _sync_file().
         """
+        now = time.time()
         cursor = self.con.execute("SELECT path, mtime FROM Files WHERE is_dir = 1")
         self._update_cursor = False
         dir = cursor.fetchone()
@@ -170,6 +315,8 @@ class FileIdManager(LoggingConfigurable):
 
             dir = cursor.fetchone()
 
+        self._last_sync = now
+
     def _sync_dir(self, dir_path):
         """
         Syncs the contents of a directory. If a child directory is dirty because
@@ -361,8 +508,7 @@ class FileIdManager(LoggingConfigurable):
     def sync_all(self):
         """
         Syncs Files table with the filesystem and ensures that the correct path
-        is associated with each file ID. Required to be called manually
-        beforehand if `sync=False` is passed to `get_path()`.
+        is associated with each file ID.
 
         Notes
         -----
@@ -408,25 +554,22 @@ class FileIdManager(LoggingConfigurable):
         self.con.commit()
         return id
 
-    def get_path(self, id, sync=True):
+    def get_path(self, id):
         """Retrieves the file path associated with a file ID. The file path is
         relative to `self.root_dir`. Returns None if the ID does not
         exist in the Files table, if the path no longer has a
         file, or if the path is not a child of `self.root_dir`.
 
-        Parameters
-        ----------
-        sync : bool
-            Whether to invoke `sync_all()` automatically prior to ID lookup.
-
         Notes
         -----
-        - For performance reasons, if this method must be called multiple times
-        in serial, then it is best to first invoke `sync_all()` and then call
-        this method with the argument `sync=False`.
+        - To force syncing when calling `get_path()`, call `sync_all()` manually
+        prior to calling `get_path()`.
         """
-        if sync:
-            self.sync_all()
+        if (
+            self.autosync_interval >= 0
+            and (time.time() - self._last_sync) >= self.autosync_interval
+        ):
+            self._sync_all()
 
         row = self.con.execute("SELECT path, ino FROM Files WHERE id = ?", (id,)).fetchone()
 
@@ -586,8 +729,17 @@ class FileIdManager(LoggingConfigurable):
         self._update(id, stat_info)
         self.con.commit()
 
+    def get_handlers_by_action(self) -> Dict[str, Optional[Callable[[Dict[str, Any]], Any]]]:
+        return {
+            "get": None,
+            "save": lambda data: self.save(data["path"]),
+            "rename": lambda data: self.move(data["source_path"], data["path"]),
+            "copy": lambda data: self.copy(data["source_path"], data["path"]),
+            "delete": lambda data: self.delete(data["path"]),
+        }
+
     def __del__(self):
-        """Cleans up `FileIdManager` by committing any pending transactions and
+        """Cleans up `LocalFileIdManager` by committing any pending transactions and
         closing the connection."""
         if hasattr(self, "con"):
             self.con.commit()
